/*
 * $Id: frame.d,v 1.2 2006/11/23 02:29:43 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.Rand;
import abagames.util.Vector2;
import abagames.util.sdl.Pad;
import abagames.util.sdl.Pad.RecordablePad;

/**
 * Game frame and actor pools.
 */
public class Frame extends abagames.util.sdl.Frame {
	static final String LAST_REPLAY_FILE_NAME = "last.rpl";
	Pad pad;
	Screen screen;
	Field field;
	Player player;
	PlayerSpec playerSpec;
	EnemyPool enemies;
	BulletPool bullets;
	ParticlePool particles;
	PillarPool pillars;
	Stage stage;
	Title title;
	Preference preference;
	GameState gameState;
	ReplayData replayData;
	Rand rand;

	public void init(GL gl) {
		Sound.load();
		preference = (Preference) abstractPreference;
		preference.load();
		Letter.init(gl);
		pad = (Pad) abstractInput;
		pad.openJoystick();
		screen = (Screen) abstractScreen;
		field = new Field(this, screen);
		enemies = new EnemyPool();
		enemies.field(field);
		bullets = new BulletPool();
		particles = new ParticlePool();
		pillars = new PillarPool();
		enemies.init(128, null);
		bullets.init(512, null);
		TriangleParticleSpec triangleParticleSpec = new TriangleParticleSpec(gl, field);
		particles.init(1024, new Object[]{triangleParticleSpec, new LineParticleSpec(field),
				new QuadParticleSpec(field), new BonusParticleSpec(field)});
		triangleParticleSpec.setParticles(particles);
		pillars.init(48, null);
		gameState = new GameState(gl, this, preference);
		title = new Title(preference, pad, this);
		title.init();
		playerSpec = new PlayerSpec(gl, pad, gameState, field, enemies, bullets, particles);
		player = new Player(gl, playerSpec);
		stage = new Stage(gl, field, enemies, bullets, player, particles, pillars, gameState);
		gameState.setStage(stage);
		rand = new Rand();
		loadLastReplay();
	}

	public void quit(GL gl) {
		title.close();
		playerSpec.close(gl);
		gameState.close(gl);
		stage.close(gl);
		Letter.close(gl);
	}

	public void start(GL gl) {
		startTitle();
	}

	public void startInGame() {
		gameState.startInGame();
		player.replayMode(false);
		RecordablePad rp = (RecordablePad) pad;
		rp.input.startRecord();
		replayData = new ReplayData();
		replayData.inputRecord = rp.input.inputRecord;
		replayData.seed = rand.nextInt32();
		clearAll();
		field.set();
		player.set();
		stage.start(replayData.seed);
		Sound.clearMarkedSes();
		Sound.playBgm();
	}

	public void startTitle() {
		startReplay();
		title.start();
	}

	public void startReplay() {
		gameState.startTitle();
		if (replayData != null && replayData.inputRecord != null) {
			player.replayMode(true);
			RecordablePad rp = (RecordablePad) pad;
			rp.input.startReplay(replayData.inputRecord);
		}
		clearAll();
		field.set();
		if (replayData != null && replayData.inputRecord != null) {
			player.set();
			stage.start(replayData.seed);
			gameState.inReplay(true);
		} else {
			field.setEyePos(new Vector2(0, 0));
		}
		Sound.clearMarkedSes();
		Sound.haltBgm();
	}

	private void clearAll() {
		enemies.clear();
		bullets.clear();
		particles.clear();
		pillars.clear();
	}

	public void breakLoop() {
		mainLoop.breakLoop();
	}

	public void move(GL gl) {
		gameState.move();
		field.move();
		if (gameState.isInGame() || (replayData != null && replayData.inputRecord != null)) {
			if (!gameState.paused()) {
				stage.move();
				pillars.move();
				player.move();
				enemies.move();
				bullets.move();
				particles.move();
			}
		}
		if (gameState.isTitle())
			title.move();
	}

	public void handleSound() {
		Sound.playMarkedSes();
	}

	public void addSlowdownRatio(float sr) {
		mainLoop.addSlowdownRatio(sr);
	}

	public void draw(GL gl) {
//		SDL_Event e = mainLoop.event;//FIXME
//		if (e.type == SDL_VIDEORESIZE) {
//			SDL_ResizeEvent re = e.resize;
//			if (re.w > 150 && re.h > 100)
//				screen.resized(re.w, re.h);
//		}
		field.setLookAt(gl);
		if (gameState.isInGame() || (replayData != null && replayData.inputRecord != null)) {
//			long t0 = System.nanoTime();
			pillars.drawOutside(gl);//.1
//			long t1 = System.nanoTime();
			field.drawBack(gl);//7
//			long t2 = System.nanoTime();
			enemies.drawPillarBack(gl);//.009
//			long t3 = System.nanoTime();
			pillars.drawCenter(gl);//.007
//			long t4 = System.nanoTime();
			enemies.drawBack(gl);//.004
//			long t5 = System.nanoTime();
			field.drawFront(gl);//4.9
//			long t6 = System.nanoTime();
			particles.draw(gl);//1.0
			enemies.drawFront(gl);
			player.draw(gl);
			bullets.draw(gl);
			field.beginDrawingFront(gl);
			gameState.draw(gl);
			if (gameState.isTitle())
				title.draw(gl);
			field.resetLookAt(gl);
			gameState.drawLeft(gl);
//			long t7 = System.nanoTime();
//			System.out.println("Frame.draw " + (t1 - t0) / 1000000.0 + " " + (t2 - t1) / 1000000.0 + " " + (t3 - t2) / 1000000.0
//			+ " " + (t4 - t3) / 1000000.0 + " " + (t5 - t4) / 1000000.0 + " " + (t6 - t5) / 1000000.0 + " " + (t7 - t6) / 1000000.0);
			//0.094984 7.205664 0.009499 0.077943 0.004749 4.974934 1.098184
			//0.094984 7.270757 0.011733 0.11454 0.007822 5.128864 1.551594
			//0.096381 6.183188 0.007264 0.058946 0.005028 4.577398 1.422247 glDrawArray
			//0.073753 4.390502 0.049168 0.036317 0.271543 3.190909 1.775086 simplify
		} else {
			pillars.drawOutside(gl);
			field.drawBack(gl);
			field.drawFront(gl);
			field.beginDrawingFront(gl);
			if (gameState.isTitle())
				title.draw(gl);
		}
	}

	public int[] keys() {
		return pad.keys;
	}

	// Handle a replay data.
	public void saveLastReplay() {
		try {
			replayData.score = gameState.score;
			saveReplay(LAST_REPLAY_FILE_NAME);
		} catch (Exception o) {
			
		}
	}

	public void loadLastReplay() {
		try {
			loadReplay(LAST_REPLAY_FILE_NAME);
			gameState.lastGameScore(replayData.score);
		} catch (Exception o) {
			resetReplay();
		}
	}

	public void saveReplay(String fileName) {
		replayData.save(fileName);
	}

	public void loadReplay(String fileName) {
		replayData = new ReplayData();
		replayData.load(fileName);
	}

	public void resetReplay() {
		replayData = null;
	}
}

class GameState {
	static final int TITLE = 0;
	static final int IN_GAME = 1;
	
	static final int EXTEND_SCORE = 100000;
	static final int MAX_LEFT = 4;
	Frame frame;
	Preference preference;
	int scene;
	Stage stage;
	int score;
	int _lastGameScore;
	int nextExtendScore;
	float _multiplier;
	int left;
	boolean escPressed, pPressed;
	boolean _paused;
	int pauseCnt;
	boolean _isGameOver;
	int gameOverCnt;
	PlayerShape playerShape;
	PlayerLineShape playerLineShape;
	boolean _inReplay;

//	invariant {
//	assert(_multiplier >= 1.0f);
//	}

	public GameState(GL gl, Frame frame, Preference preference) {
		this.frame = frame;
		this.preference = preference;
		playerShape = new PlayerShape(gl);
		playerLineShape = new PlayerLineShape(gl);
		clear();
		_lastGameScore = -1;
	}

	public void setStage(Stage stage) {
		this.stage = stage;
	}

	public void close(GL gl) {
		playerShape.close(gl);
		playerLineShape.close(gl);
	}

	public void startInGame() {
		scene = IN_GAME;
		clear();
		left = 2;
		_lastGameScore = -1;
	}

	public void startTitle() {
		scene = TITLE;
		clear();
		left = 2;
	}

	private void clear() {
		score = 0;
		nextExtendScore = EXTEND_SCORE;
		_multiplier = 1.0f;
		left = 0;
		gameOverCnt = 0;
		_isGameOver = _paused = false;
		_inReplay = false;
	}

	public void startGameOver() {
		if (!isInGameAndNotGameOver())
			return;
		_isGameOver = true;
		gameOverCnt = 0;
		Sound.fadeBgm();
		_lastGameScore = score;
		preference.recordResult(score);
		preference.save();
	}

	public void startGameOverWithoutRecording() {
		if (_isGameOver)
			return;
		_isGameOver = true;
		gameOverCnt = 0;
		Sound.fadeBgm();
	}

	public void backToTitle() {
		if (isTitle()) {
			frame.startReplay();
			return;
		}
		if (gameOverCnt > 120) {
			frame.saveLastReplay();
			frame.startTitle();
		}
	}

	public void move() {
		handleEscKey();
		if (isInGameAndNotGameOver()) {
			handlePauseKey();
			if (_paused) {
				pauseCnt++;
				return;
			}
		}
		if (isInGame()) {
			if (!_isGameOver) {
				frame.handleSound();
			} else {
				gameOverCnt++;
				if (gameOverCnt < 60)
					frame.handleSound();
				if (gameOverCnt > 1000)
					backToTitle();
			}
		} else {
			if (_inReplay)
				frame.handleSound();
			if (_isGameOver) {
				gameOverCnt++;
				if (_inReplay && gameOverCnt < 60)
					frame.handleSound();
				if (gameOverCnt > 120)
					backToTitle();
			}
		}
	}

	private void handleEscKey() {
//		if (frame.keys()[SDLK_ESCAPE] == SDL_PRESSED) {//FIXME
		if (frame.keys()[Pad.SDLK_ESCAPE] == Pad.SDL_PRESSED) {
			if (!escPressed) {
				escPressed = true;
				if (scene == IN_GAME) {
					frame.loadLastReplay();
					frame.startTitle();
				} else {
					frame.breakLoop();
				}
			}
		} else {
			escPressed = false;
		}
	}

	private void handlePauseKey() {
		if (frame.keys()[Pad.SDLK_p] == Pad.SDL_PRESSED) {//FIXME
//		if (false) {
			if (!pPressed) {
				pPressed = true;
				_paused = !_paused;
				pauseCnt = 0;
			}
		} else {
			pPressed = false;
		}
	}

	public void addScore(int sc) {
		addScore(sc, false);
	}
	public void addScore(int sc, boolean noMultiplier) {
		if (!_isGameOver) {
			if (noMultiplier)
				score += sc;
			else
				score += sc * _multiplier;
			if (score >= nextExtendScore) {
				if (left < MAX_LEFT) {
					left++;
					Sound.playSe("extend.wav");
				}
				nextExtendScore += EXTEND_SCORE;
			}
		}
	}

	public void addMultiplier(float mp) {
		if (!_isGameOver)
			_multiplier += mp;
	}

	public void mulMultiplier(float mp) {
		if (!_isGameOver) {
			_multiplier *= mp;
			if (_multiplier < 1)
				_multiplier = 1;
		}
	}

	public void destroyedPlayer() {
		left--;
		if (left < 0) {
			if (isInGame())
				startGameOver();
			else
				startGameOverWithoutRecording();
		}
	}

	public void countShotFired() {
		stage.countShotFired();
	}

	public void countShotHit() {
		stage.countShotHit();
	}

	public void draw(GL gl) {
		Letter.drawNum(gl, score, 132, 5, 7);
		if (_lastGameScore >= 0)
			Letter.drawNum(gl, _lastGameScore, 360, 5, 7);
		Letter.drawNum(gl, (int) (_multiplier * 100), 626, 4, 9, 3, 33, 2);
		stage.drawPhaseNum(gl);
		if (isInGame()) {
			if (!_isGameOver)
				stage.draw(gl);
			if (_isGameOver) {
				if (gameOverCnt > 60) {
					Letter.drawString(gl, "GAME OVER", 214, 200, 12);
					stage.drawGameover(gl);
				}
			} else if (_paused) {
				if (pauseCnt % 120 < 60)
					Letter.drawString(gl, "PAUSE", 290, 420, 7);
			}
		}
	}

	public void drawLeft(GL gl) {
		for (int i = 0; i < left; i++) {
			gl.glPushMatrix();
			gl.glTranslatef(-10.2f + i, -7.5f, -10);
			gl.glScalef(0.6f, 0.6f, 0.6f);
			playerShape.draw(gl);
			Screen.setColor(gl, 0, 0, 0);
			playerLineShape.draw(gl);
			gl.glPopMatrix();
		}
	}

	public boolean isInGame() {
		return (scene == IN_GAME);
	}

	public boolean isInGameAndNotGameOver() {
		return (scene == IN_GAME && !_isGameOver);
	}

	public boolean isTitle() {
		return (scene == TITLE);
	}

	public boolean isGameOver() {
		return _isGameOver;
	}

	public boolean paused() {
		return _paused;
	}

	public float multiplier() {
		return _multiplier;
	}

	public boolean inReplay(boolean v) {
		return _inReplay = v;
	}

	public int lastGameScore(int v) {
		return _lastGameScore = v;
	}
}
